/* 
linux/fs/dmsdos/dmsdos_alloc.c

DMSDOS filesystem: cluster allocation functions

******************************************************************************
DMSDOS (Doublespace/Drivespace compressed MSDOS filesystem) for Linux
written 1995,1996 by Frank Gockel

    (C) Copyright 1995-1997 by Frank Gockel
    (C) Copyright 1996,1997 by Pavel Pisa (stacker access)
        
Some code of the dmsdos filesystem has been copied from the msdos filesystem
so there are the following additional copyrights:

    (C) Copyright 1992,1993 by Werner Almesberger (msdos filesystem)
    (C) Copyright 1994,1995 by Jacques Gelinas (mmap code)
    (C) Copyright 1992-1995 by Linus Torvalds

The DMSDOS filesystem was inspired by the THS filesystem (a simple doublespace
DS-0-2 compressed read-only filesystem) written 1994 by Thomas Scheuermann.

The DMSDOS filesystem is distributed under the Gnu General Public Licence.
See file COPYING for details.
******************************************************************************

*/

#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/msdos_fs.h>
#include <linux/string.h>
#include <linux/sched.h>
#include "dmsdos.h"

extern unsigned long dmsdos_speedup;

#define NEAR_AREA 512
/* no change for doublespace, but fixes drivespace 3 problems (was 50) */
#define BIG_HOLE (dblsb->s_sectperclust*3+2) 

/* experimental new memory allocation routines */
#ifdef USE_XMALLOC

int told=0;

#include<linux/mm.h>
#include<linux/malloc.h>

#ifndef MAX_KMALLOC_SIZE
#define MAX_KMALLOC_SIZE 128*1024
#endif

void* xmalloc(unsigned long size)
{ void* result;

  if(size<=MAX_KMALLOC_SIZE)
  { result=kmalloc(size,GFP_KERNEL);
    if(result)
    { /* the xfree routine recognizes kmalloc'd memory due to that it is
         usually not page-aligned --  we double-check here to be sure */
      if (  ( ((unsigned long)result) & (PAGE_SIZE-1) ) ==0   )
      { /* uhhh.... the trick is broken... 
           tell the user and go safe using vmalloc */
        kfree(result);
        if(told++)printk("DMSDOS: page-aligned memory returned by kmalloc - please disable XMALLOC\n"); 
      }
      else return result;
    }
  }
  result=vmalloc(size);
  /* again check alignment to be 100% sure */
  if ( ((unsigned long)result) & (PAGE_SIZE-1) )
    panic("DMSDOS: vmalloc returned unaligned memory - please disable XMALLOC\n"); 
  return result;
}

void xfree(void * data)
{ unsigned long address=(unsigned long)data;

  /* we rely on the fact that kmalloc'ed memory is unaligned but
     vmalloc'ed memory is page-aligned */
  /* we are causing SERIOUS kernel problems if the wrong routine is called -
     therefore xmalloc tests kmalloc's return address above */
  if(address&(PAGE_SIZE-1))kfree(data);
  else vfree(data);
}
#endif

#ifndef NEW_ACCESS_LOCK
struct wait_queue * fat_alloc_wait = NULL;
int fat_alloc_lock=0;
void lock_fat_alloc(void)
{ while(fat_alloc_lock)sleep_on(&fat_alloc_wait);
  fat_alloc_lock=1;
}
void unlock_fat_alloc(void)
{ fat_alloc_lock=0;
  wake_up(&fat_alloc_wait);
}

struct wait_queue * mdfat_alloc_wait = NULL;
int mdfat_alloc_lock=0;
void lock_mdfat_alloc(void)
{ while(mdfat_alloc_lock)sleep_on(&mdfat_alloc_wait);
  mdfat_alloc_lock=1;
}
void unlock_mdfat_alloc(void)
{ mdfat_alloc_lock=0;
  wake_up(&mdfat_alloc_wait);
}

#else  /* ifndef NEW_ACCESS_LOCK / else */

void lock_fat_alloc(Dblsb*dblsb)
{ while(dblsb->fat_alloc_locked)sleep_on(&(dblsb->fat_alloc_wait));
  dblsb->fat_alloc_locked=1;
}
void unlock_fat_alloc(Dblsb*dblsb)
{ dblsb->fat_alloc_locked=0;
  wake_up(&(dblsb->fat_alloc_wait));
}

void lock_mdfat_alloc(Dblsb*dblsb)
{ while(dblsb->mdfat_alloc_locked)sleep_on(&(dblsb->mdfat_alloc_wait));
  dblsb->mdfat_alloc_locked=1;
}
void unlock_mdfat_alloc(Dblsb*dblsb)
{ dblsb->mdfat_alloc_locked=0;
  wake_up(&(dblsb->mdfat_alloc_wait));
}

#endif  /* ifndef NEW_ACCESS_LOCK / else */

void u_free_cluster_sectors(struct super_block*sb, int clusternr,
                            unsigned long* undo_list)
{ 
  Mdfat_entry mde,dummy,newmde;
  int newval=0;
  int i;
  int sektors;
  int sektornr;
  int undo_pnt=0;
  Dblsb*dblsb=MSDOS_SB(sb)->private_data;
  
  LOG_ALLOC("DMSDOS: free_cluster_sectors: freeing cluster %d\n",clusternr);

  /* read mdfat entry and clear */    
  newmde.sector_minus_1=0;
  newmde.size_lo_minus_1=0;
  newmde.size_hi_minus_1=0;
  newmde.flags=0;
  dbl_mdfat_value(sb,clusternr,NULL,&mde);
  dbl_mdfat_value(sb,clusternr,&newmde,&dummy);
  sektors=mde.size_lo_minus_1+1;
  sektornr=mde.sector_minus_1+1;
  if(mde.flags&2)
  { if(mde.unknown&2)
    { /* free fragmented cluster */
      struct buffer_head*bh;
      int fragcount;
      int fragpnt;
      int sec;
      int cnt;
      
      bh=raw_bread(sb,sektornr);
      if(bh==NULL)
      { printk("DMSDOS: free_cluster_sectors: fragmentation list unreadable in cluster %d\n",
               clusternr);
        goto nfree;
      }
      fragcount=bh->b_data[0];
      if(fragcount<=0||fragcount>dblsb->s_sectperclust||
         bh->b_data[1]!=0||bh->b_data[2]!=0||bh->b_data[3]!=0)
      { printk("DMSDOS: free_cluster_sectors: error in fragmentation list in cluster %d\n",
               clusternr);
        raw_brelse(sb,bh);
        goto nfree;
      }
      for(fragpnt=1;fragpnt<=fragcount;++fragpnt)
      { cnt=bh->b_data[fragpnt*4+3];
        cnt&=0xff;
        cnt/=4;
        cnt+=1;
        sec=bh->b_data[fragpnt*4];
        sec&=0xff;
        sec+=bh->b_data[fragpnt*4+1]<<8;
        sec&=0xffff;
        sec+=bh->b_data[fragpnt*4+2]<<16;
        sec&=0xffffff;
        sec+=1;
        
        if(fragpnt==1)
        { if(sec!=sektornr||cnt!=sektors)
          { printk("DMSDOS: free_cluster_sectors: first fragment wrong in cluster %d\n",
                   clusternr);
            raw_brelse(sb,bh);
            goto nfree;
          }
        }
        
        for(i=0;i<cnt;++i)
        { dbl_bitfat_value(sb,sec+i,&newval);
          if(undo_list)undo_list[undo_pnt++]=sec+i;
        }
      }
    }
    else
    { /* free sectors in bitfat */
     nfree:
      for(i=0;i<sektors;++i)
      { dbl_bitfat_value(sb,sektornr+i,&newval);
        if(undo_list)undo_list[undo_pnt++]=sektornr+i;
      }
    }
    
    dblsb->s_full=0;
  }
  else
  { LOG_CLUST("DMSDOS: stale MDFAT entry for cluster %d, zeroing.\n",
           clusternr);
  }
  
  if(undo_list)undo_list[undo_pnt]=0;
}

void free_cluster_sectors(struct super_block*sb, int clusternr)
{ lock_mdfat_alloc();
  u_free_cluster_sectors(sb,clusternr,NULL);
  unlock_mdfat_alloc();
}

/* for statistics - just for interest */
int nearfound=0;
int bigfound=0;
int exactfound=0;
int anyfound=0;
int notfound=0;

/* this function must be called locked */
int find_free_bitfat(struct super_block*sb, int sektornr, int size)
{ int testsek;
  int i;
  Dblsb*dblsb=MSDOS_SB(sb)->private_data;

if(dmsdos_speedup&SP_FAST_BITFAT_ALLOC)
{ /* new strategy: find any fitting hole beginning with last result */

  testsek=dblsb->s_lastnear;
  if(testsek<dblsb->s_datastart||testsek>dblsb->s_dataend-size)
    testsek=dblsb->s_datastart;
    
  while(testsek<=dblsb->s_dataend-size)
  { if(dbl_bitfat_value(sb,testsek,NULL))
    { ++testsek;
      continue;
    }
    i=1;
    while(i<size&&dbl_bitfat_value(sb,testsek+i,NULL)==0)++i;
    if(i==size)
    { ++nearfound;
      dblsb->s_lastnear=testsek+size;
      dblsb->s_full=0;
      return testsek;
    }
    testsek+=i;
  }
  
  /* not found, continue */
  goto tryany;
  
} /* end of new strategy */
else
{ /* old strategy: do a search in near environment first */

  if(sektornr==0)sektornr=dblsb->s_lastnear;

  if(sektornr>=dblsb->s_datastart&&sektornr<=dblsb->s_dataend-size)
  { /* search exactly fitting hole near sektornr */
    testsek=sektornr;  
    while(testsek<sektornr+NEAR_AREA)
    { if(dbl_bitfat_value(sb,testsek,NULL))
      { ++testsek;
        continue;
      }
      i=1;
      while(i<=size&&dbl_bitfat_value(sb,testsek+i,NULL)==0)++i;
      if(i==size)
      { dblsb->s_full=0;
        ++nearfound;
        dblsb->s_lastnear=testsek; 
        return testsek;
      }
      testsek+=i;
    }
    testsek=sektornr;
    while(testsek>sektornr-NEAR_AREA)
    { if(dbl_bitfat_value(sb,testsek,NULL))
      { --testsek;
        continue;
      }
      i=1;
      while(i<=size&&dbl_bitfat_value(sb,testsek-i,NULL)==0)++i;
      if(i==size)
      { dblsb->s_full=0;
        ++nearfound;
        dblsb->s_lastnear=testsek-i+1;
        return testsek-i+1;
      }
      testsek-=i;
    }
  }
  /* not found, continue */
} /* end of old strategy */

  /* search for a big hole */
  if(dblsb->s_lastbig==-1)goto nobighole;
  
  testsek=dblsb->s_lastbig;
  if(testsek<dblsb->s_datastart||testsek+size>dblsb->s_dataend)
    testsek=dblsb->s_datastart;
    
  while(testsek<=dblsb->s_dataend-size)
  { if(dbl_bitfat_value(sb,testsek,NULL))
    { ++testsek;
      continue;
    }
    i=1;
    while(i<BIG_HOLE&&dbl_bitfat_value(sb,testsek+i,NULL)==0)++i;
    if(i==BIG_HOLE)
    { dblsb->s_full=0;
      ++bigfound;
      dblsb->s_lastbig=testsek;
      return testsek;
    }
    testsek+=i;
  }
  
  if(dblsb->s_lastbig==0)
     dblsb->s_lastbig=-1; /*there's no big hole any more*/
  else
     dblsb->s_lastbig=0; /* next time try from the beginning */

nobighole:

if((dmsdos_speedup&SP_NO_EXACT_SEARCH)==0)  
{
  /* search for an exactly fitting hole */
  /* hmmm... now the search code becomes awfully slow */
  testsek=dblsb->s_datastart;
  while(testsek<=dblsb->s_dataend-size)
  { if(dbl_bitfat_value(sb,testsek,NULL))
    { ++testsek;
      continue;
    }
    i=1;
    while(i<=size&&dbl_bitfat_value(sb,testsek+i,NULL)==0)++i;
    if(i==size)
    { dblsb->s_full=0;
      ++exactfound;
      return testsek;
    }
    testsek+=i;
  }
}
  
  if(dblsb->s_full==0)
  printk(KERN_EMERG "DMSDOS: CVF almost full or highly fragmented at MDFAT level, refusing further non-root write access.\n");
  
  dblsb->s_full=1;

tryany:
  /* last trial: search for any hole >= size */
  testsek=dblsb->s_datastart;
  while(testsek<=dblsb->s_dataend-size)
  { if(dbl_bitfat_value(sb,testsek,NULL))
    { ++testsek;
      continue;
    }
    i=1;
    while(i<size&&dbl_bitfat_value(sb,testsek+i,NULL)==0)++i;
    if(i==size)
    { ++anyfound;
      return testsek;
    }
    testsek+=i;
  }
  
  /* not found, means disk full or MDFAT too fragmented */
  ++notfound;
  dblsb->s_full=2;
  printk(KERN_EMERG "DMSDOS: CVF full or too fragmented at MDFAT level\n");
  return 0;
}

void log_found_statistics()
{ printk("DMSDOS free sector finding statistics:\n");
  printk("nearfound=%d bigfound=%d exactfound=%d anyfound=%d notfound=%d\n",
         nearfound,bigfound,exactfound,anyfound,notfound);
}


/* replaces an existing cluster;
   this unusual function must be called before rewriting any file cluster;
   *** size must be known (encoded in mde) ***
   it does nothing if called too often;
   returns first sector nr
*/

#define MAX_UNDO_LIST 70

int replace_existing_cluster(struct super_block*sb, int cluster, 
                             int near_sector,
                             Mdfat_entry*mde)
{ Mdfat_entry old_mde,new_mde,dummy;
  int i;
  int newval;
  int sektor;
  int old_sektor;
  int old_size;
  int new_size;
  unsigned long undo_list[MAX_UNDO_LIST];
  Dblsb*dblsb=MSDOS_SB(sb)->private_data;

  lock_mdfat_alloc();

  LOG_ALLOC("DMSDOS: replace_existing_cluster cluster=%d near_sector=%d\n",
         cluster,near_sector);
  dbl_mdfat_value(sb,cluster,NULL,&old_mde);
  old_size=old_mde.size_lo_minus_1+1;
  old_sektor=old_mde.sector_minus_1+1;
  new_size=mde->size_lo_minus_1+1;
  if(old_mde.flags&2)
  {
    /* test whether same length (and not fragmented in drivespace 3) */
    if(old_size==new_size&&
       (dblsb->s_cvf_version!=DRVSP3||(mde->unknown&2)==0)&&
       (dblsb->s_cvf_version<STAC3) /* hack */
      )
    { LOG_ALLOC("DMSDOS: replace_existing_cluster: same length, ok\n");
      sektor=old_sektor;
      goto mdfat_update;
    }
    /* different length, replace mdfat entry */
    newval=0;
    LOG_ALLOC("DMSDOS: replace_existing_cluster: freeing old sectors...\n");
    /* !!!!!! hack for Stacker to work, must be moved to
              stacker_alloc in future    Pavel */
    if(dblsb->s_cvf_version>=STAC3)
      stac_special_free(sb,cluster);
    else   
      /*for(i=0;i<old_size;++i)dbl_bitfat_value(sb,old_sektor+i,&newval);*/
      u_free_cluster_sectors(sb,cluster,undo_list);
    LOG_ALLOC("DMSDOS: replace_existing_cluster: freeing finished\n");
  }
  LOG_ALLOC("DMSDOS: replace_existing_cluster: call find_free_bitfat...\n");
  sektor=find_free_bitfat(sb,near_sector,new_size);
  LOG_ALLOC("DMSDOS: replace_existing_cluster: find_free_bitfat returned %d\n",
         sektor);
  if(sektor==0)
  { if(old_mde.flags&2)
    { if(dblsb->s_cvf_version<=DRVSP3)
      { /* undo bitfat free */
        LOG_ALLOC("DMSDOS: replace_existing_cluster: undoing bitfat free...\n");
        newval=1;
        for(i=0;undo_list[i]!=0;++i)
               dbl_bitfat_value(sb,undo_list[i],&newval);
      }
      else
      { /* Stacker routines don't have an undo list.
           We cannot restore the state before. Sorry data are lost now. */
        new_mde.sector_minus_1=0;
        new_mde.size_lo_minus_1=0;
        new_mde.size_hi_minus_1=0;
        new_mde.flags=mde->flags=0;
        LOG_ALLOC("DMSDOS: replace_existing_cluster: deleting mdfat entry...\n");
        dbl_mdfat_value(sb,cluster,&new_mde,&dummy);
      }
    }
    unlock_mdfat_alloc();
    return -ENOSPC; /* disk full */
  }
  /* check whether really free (bug supposed in find_free_bitfat) */
  for(i=0;i<new_size;++i)
  { if(dbl_bitfat_value(sb,sektor+i,NULL))
    { printk(KERN_EMERG "DMSDOS: find_free_bitfat returned sektor %d size %d but they are not all free!\n",
             sektor,new_size);
      unlock_mdfat_alloc();
      panic("This is a dmsdos bug - reboot and check filesystem\n");
      return -EIO;
    }
  }
  newval=1;
  LOG_ALLOC("DMSDOS: replace_existing_cluster: allocating in bitfat...\n");
  for(i=0;i<new_size;++i)dbl_bitfat_value(sb,sektor+i,&newval);
  
mdfat_update:
  new_mde.sector_minus_1=sektor-1;
  new_mde.size_lo_minus_1=mde->size_lo_minus_1;
  new_mde.size_hi_minus_1=mde->size_hi_minus_1;
  new_mde.flags=mde->flags|2;
  LOG_ALLOC("DMSDOS: replace_existing_cluster: writing mdfat...\n");
  dbl_mdfat_value(sb,cluster,&new_mde,&dummy);
  unlock_mdfat_alloc();
  return sektor; /* okay */
}

